import { useEffect, useState } from 'react'

import { SetState } from '../utils/types'
import useSend from './useSend'

type UseFetchShared<T> = {
    loading: boolean,
    abort?: () => void,
    setData: SetState<T | undefined>
}

type UseFetchSucced<T> = {
    error: null,
    data: T,
} & UseFetchShared<T>

type UseFetchErrored<T> = {
    error: string,
    data: undefined
} & UseFetchShared<T>

const gotError = <T>(res: UseFetchErrored<T> | UseFetchSucced<T>): res is UseFetchErrored<T> => (
    typeof res.error === 'string'
)

const fallbackError = <T>(res: UseFetchSucced<T> | UseFetchErrored<T>) => (
    gotError(res) ? res.error : res.data
)

type UseFetchReturn<T> = UseFetchSucced<T> | UseFetchErrored<T>

function useFetch<R, T>(
    url: string,
    method: RequestInit['method'],
    needAuth: boolean,
    guardResponse: (data: unknown) => data is R,
    processResponse: (data: R) => T,
    initialData?: T,
    params?: Omit<RequestInit, 'method'>
): UseFetchReturn<T> {
    const [data, setData] = useState(initialData)

    const { doSend, loading, error } = useSend(url, method, needAuth, guardResponse, processResponse, params)

    useEffect(() => {
        doSend().then(
            data => { if (data !== undefined) setData(data) }
        ).catch( // must never occur
            err => import.meta.env.DEV && console.error('Failed to do fetch request', err)
        )
    }, [doSend])

    return {
        ...(
            error === null ? ({
                data: data!, error: null
            }) : ({ data: undefined, error })
        ),
        loading,
        setData
    }
}

export type { UseFetchReturn }

export default useFetch

export { gotError, fallbackError }
